/*
 * PCB Exposure Timer, main.c
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 193 2012-10-30 23:04:21Z schwenke $
 *
 */

/* configurables */

#define DEFAULT_TIME 120    /*  2:00 min */
#define MAXIMUM_TIME 30*60  /* 30:00 min */

/* encoder acceleration characteristics */
#define ACC_UP 16
#define ACC_SCALE 32
#define ACC_MAX 20

/* end configurables */


#if MAXIMUM_TIME > 99*60+59
#error MAXIMUM time exceeds 99:59 min
#endif


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include <util/atomic.h>
#include <util/delay.h>
#include "hal.h"


/* global variables */
struct
{
    int16_t time;  /* seconds */
    uint8_t state; /* enum, see below */
} glob;

/* possible values for glob.state */
#define READY   0
#define RUNNING 1
#define DONE    2

/* display buffer, 5*8 bits, first 35 bits used */
uint8_t display[5];

/* 7 segment decoding table
 * 1=a, 2=b, 4=c ... 0x40=f */
const uint8_t n_to_7[10] =
{
    0x3F, 0x06, 0x5B, 0x4F, 0x66, /* 0-4 */
    0x6D, 0x7D, 0x07, 0x7F, 0x6F  /* 5-9 */
};

/* rotary encoder decoding table */
const uint8_t enc_table[16] =
{
    0,  0, -1, 0,
    0,  0,  0, 1,
    1,  0,  0, 0,
    0, -1,  0, 0
};


/* clock tick, incremented every 1 ms */
volatile uint16_t vtick;

/* encoder status */
volatile int8_t  enc_delta;  /* -128..+127 */
volatile uint8_t enc_button; /* enum, see below */
#define pressed_short 1
#define pressed_long  2

/* default timer interval in EEPROM */
EEMEM uint16_t def_time = DEFAULT_TIME;


/* function prototypes */
void debounce_button(void);
void debounce_encoder(void);
void hw_init(void);
void refresh(void);
void restart_timer(void);
void setup_display(void);


/* here we go! */
int __attribute__((OS_main)) main (void)
{
    hw_init();
    glob.time = eeprom_read_word(&def_time);
    glob.state = READY;
    setup_display();
    refresh();

    while (1) {

        uint8_t  need_refresh = 0;
        uint8_t  button;
        int8_t   delta;
        uint16_t tick;
        /* self init avoids bogus 'uninitialized' warning */
        int16_t  last_time = last_time;

        /* wait for next timer interrupt */
        sleep_cpu();

        /* read volatile variables */
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            button = enc_button; enc_button = 0;
            delta = enc_delta; enc_delta = 0;
            tick = vtick;
        }

        switch (glob.state) {

        case READY:

            if (delta) {
                glob.time += delta;
                if (glob.time < 0) {
                    glob.time = 0;
                }
                if (glob.time > MAXIMUM_TIME) {
                    glob.time = MAXIMUM_TIME;
                }
                need_refresh = 1;
            }

            if (button == pressed_short) {
                /* start timer */
                last_time = glob.time;
                glob.state = RUNNING;
                restart_timer();
                bulb_on();
                need_refresh = 1;
            }
            if (button == pressed_long) {
                /* save time in EEPROM */
                eeprom_write_word(&def_time, glob.time);
                glob.state = DONE;
                need_refresh = 1;
            }
            break;

        case RUNNING:

            if (tick == 0) {
                glob.time--;
                need_refresh = 1;
            }
            /* button press stops timer too */
            if (glob.time == 0 || button) {
                bulb_off();
                glob.state = DONE;
                need_refresh = 1;
            }
            break;

        case DONE:

            if (button) {
                if (glob.time == 0) {
                    glob.time = last_time;
                }
                glob.state = READY;
                need_refresh = 1;
            }
            break;

        }

        if (need_refresh) {
            setup_display();
            refresh();
            need_refresh = 0;
        }
    }
}


/********************************************************************
 * Sub Routines
 ********************************************************************/

/* scan and debounce the push button
 * we detect short press (at time of button release)
 * and long press (after ~1 sec) */
void debounce_button(void)
{
    static uint8_t bst;  /* button status */
    static uint8_t cnt;  /* button debounce counter */

    /* check if the button status has changed */
    uint8_t i = (bst ^ sample_button) & 1;

    if (i) {
        bst ^= i; /* remember current status  */
        cnt = 25; /* set debounce time = 25ms */
    }
    else {
        /* decrement the debounce/repeat counter */
        if (cnt && --cnt == 0) {
            /* we have a debounced or repeated action */

            if (bst & 1) { /* press */
                /* bst is filled with 1's from the right
                 * after each repeat cycle. So after 6
                 * cycles 997ms elapsed (6*162 + 25) */
                if (bst == 0x7F) {
                    enc_button = pressed_long;
                }
                /* fill bst with 1's from the right */
                bst = 2*bst+1;
                /* set repeat counter = 162ms */
                cnt = 162;
            }
            else {
                /* released after less than6 repeats? */
                if ((bst & 0x80) == 0) {
                    enc_button = pressed_short;
                }
                /* clear status and disable counter */
                bst = 0;
                cnt = 0;
            }
        }
    }
}


/* half step with acceleration
 * acc goes up by ACC_UP when encoder make a step
 * and down by 1 each ms when encoder makes no step
 * min acc is ACC_SCALE, max is ACC_SCALE*ACC_MAX
 * delta is +/- acc/ACC_SCALE */
void debounce_encoder(void)
{
    static uint8_t last; /* encoder status */
    static int16_t acc;  /* acceleration */

    last = ((last << 2) | (phase_A ? 2 : 0) | (phase_B ? 1 : 0)) & 0x0F;
    int8_t delta = enc_table[last];

    if (delta) {
        acc = (acc < ACC_MAX*ACC_SCALE-ACC_UP ? acc+ACC_UP : ACC_MAX*ACC_SCALE);
        if (delta > 0) {
            enc_delta += (acc/ACC_SCALE);
        } else {
            enc_delta -= (acc/ACC_SCALE);
        }
    } else {
        acc = (acc>ACC_SCALE ? acc-1 : ACC_SCALE);
    }
}


/* initialize the hardware
 * all unused pins are inputs w/ pullup enabled
 * timer 1 interrupts every 1 ms */
void hw_init(void)
{
    /* port B is all inputs */
    PORTB = 0xFF;
    DDRB  = 0x00;

    /* 4 outputs on port D, bulb off */
    PORTD = 0xFF & ~_BV(bulb_pin);
    DDRD  = (_BV(oe_pin) | _BV(clk_pin) | _BV(data_pin) | _BV(bulb_pin));

    /* turn off analog comparator */
    ACSR |= _BV(ACD);

    /* timer1: pre=64, count to 125 in CTC mode */
    TCCR1B = _BV(CTC1) | _BV(CS11) | _BV(CS10);
    OCR1   = 125;
    /* interrupt on compare match */
    TIMSK  = _BV(OCIE1A);

    /* enable idle sleep */
    MCUCR  = _BV(SE);

    sei();
}


/* copy display buffer onto display */
void refresh(void)
{
    /* start data frame */
    clk_lo();
    oe_lo();

    /* send start bit */
    data_hi();
    wait(); clk_hi();
    wait(); clk_lo();

    /* followed by 35 data bits */
    for (uint8_t bit=0; bit<35; bit++) {
        if (display[bit/8] & (1<<(bit%8))) {
            data_hi();
        } else {
            data_lo();
        }
        wait(); clk_hi();
        wait(); clk_lo();
    }

    /* end data frame */
    oe_hi();
}


/* restart time measurement */
void restart_timer(void)
{
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        vtick = 0;
        TCNT1 = 0;
        TIFR  = _BV(OCIE1A);
    }
}


/* setup display buffers from global variables */
void setup_display(void)
{
    /* setup digits */
    uint16_t t = (uint16_t)glob.time;
    display[3] = n_to_7[t%10]; t /= 10;
    display[2] = n_to_7[t%6];  t /= 6;
    display[1] = n_to_7[t%10]; t /= 10;
    display[0] = (t == 0 ? 0 : n_to_7[t%10]); /* suppress leading 0 */

    /* turn on colon between minutes and seconds */
    display[1]  |= 0x80;    /* bit #16 */

    /* setup status LEDs */
    display[4] = glob.state == READY ? 0x02 : glob.state == RUNNING ? 0x01 : 0;
    if (glob.state == READY) {
        display[4] = 0x02;  /* bit #34 */
    }
    if (glob.state == RUNNING) {
        display[4] = 0x01;  /* bit #33 */
    }
    if (glob.state == DONE) {
        display[0] |= 0x80; /* bit #8 */
    }
}


/* once every millisecond */
ISR(TIMER1_COMP1_vect)
{
    debounce_encoder();
    debounce_button();
    vtick = (vtick == 999 ? 0 : vtick+1);
}


